/*
 * hostapd / IEEE 802.11 Management: Beacon and Probe Request/Response
 * Copyright (c) 2002-2004, Instant802 Networks, Inc.
 * Copyright (c) 2005-2006, Devicescape Software, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */

#include "includes.h"
#include "hostapd.h"
#include "ieee802_11.h"
#include "ieee802_11n.h"

u8 * hostapd_eid_ht_capabilities_info(struct hostapd_data *hapd, u8 *eid)
{
	u8 *pos = eid;

	memcpy(pos, &hapd->conf->ht_capabilities, sizeof(struct ht_cap_ie));
 	pos += sizeof(struct ht_cap_ie);
	return pos;
}

u8 * hostapd_eid_ht_additional_info(struct hostapd_data *hapd, u8 *eid)
{
	u8 *pos = eid;

	memcpy(pos, &hapd->conf->ht_additional_info, sizeof(struct ht_add_info_ie));
	pos += sizeof(struct ht_add_info_ie);
	return pos;
}

/*
op_mode
Set to 0 (HT pure) under the followign conditions
	- all STAs in the BSS are 20/40 MHz HT in 20/40 MHz BSS or
	- all STAs in the BSS are 20 MHz HT in 20 MHz BSS
Set to 1 (HT non-member protection) if there may be non-HT STAs
	in both the primary and the secondary channel
Set to 2 if only HT STAs are associated in BSS,
	however and at least one 20 MHz HT STA is associated
Set to 3 (HT mixed mode) when one or more non-HT STAs are associated
	(currently non-GF HT station is considered as non-HT STA also)
*/
int hostapd_ht_additional_info_update(struct hostapd_data *hapd)
{
	struct ht_add_info_ie *ht_add_info;
	u16 operation_mode = 0;
	u16 cur_op_mode, new_op_mode;
	int op_mode_changes = 0;

	if (hapd == NULL)
		return -1;

	if (!hapd->conf->support11n || hapd->conf->ht_op_mode_fixed)
		return 0;

	ht_add_info = &hapd->conf->ht_additional_info;
	operation_mode = le_to_host16(ht_add_info->data.operation_mode);
	wpa_printf(MSG_DEBUG, "%s current operation mode=0x%X\n",
		   __func__, operation_mode);
	
	if ((operation_mode & HT_INFO_OPERATION_MODE_NON_GF_DEVS_PRESENT) == 0
	     && hapd->iface->num_sta_ht_no_gf) {
		operation_mode |= HT_INFO_OPERATION_MODE_NON_GF_DEVS_PRESENT;
		op_mode_changes++;
	} else if ((operation_mode & HT_INFO_OPERATION_MODE_NON_GF_DEVS_PRESENT)
	     && hapd->iface->num_sta_ht_no_gf == 0) {
		operation_mode &= ~HT_INFO_OPERATION_MODE_NON_GF_DEVS_PRESENT;
		op_mode_changes++;
	}

	if ((operation_mode & HT_INFO_OPERATION_MODE_NON_HT_STA_PRESENT) == 0
	    && (hapd->iface->num_sta_no_ht || hapd->iface->olbc_ht)) {
		operation_mode |= HT_INFO_OPERATION_MODE_NON_HT_STA_PRESENT;
		op_mode_changes++;
	} else if ((operation_mode & HT_INFO_OPERATION_MODE_NON_HT_STA_PRESENT)
	   && (hapd->iface->num_sta_no_ht == 0 && hapd->iface->olbc_ht == 0)) {
		operation_mode &= ~HT_INFO_OPERATION_MODE_NON_HT_STA_PRESENT;
		op_mode_changes++;
	}

	/* Note: currently we switch to the MIXED op mode if HT non-greenfield
	 * station is associated. Probably it's a theoretical case, since
	 * it looks like that all known HT STAs support greenfield.
	 */
	new_op_mode = 0;
	if (hapd->iface->num_sta_no_ht ||
	    (operation_mode & HT_INFO_OPERATION_MODE_NON_GF_DEVS_PRESENT))
		new_op_mode = OP_MODE_MIXED;
	else if ((hapd->conf->ht_capabilities.data.capabilities_info &
		  HT_CAP_INFO_SUPP_CHANNEL_WIDTH_SET) &&
		  hapd->iface->num_sta_ht_20mhz)
		new_op_mode = OP_MODE_20MHZ_HT_STA_ASSOCED;
	else if (hapd->iface->olbc_ht)
		new_op_mode = OP_MODE_MAY_BE_LEGACY_STAS;
	else
		new_op_mode = OP_MODE_PURE;

	cur_op_mode = operation_mode & HT_INFO_OPERATION_MODE_OP_MODE_MASK;
	if (cur_op_mode != new_op_mode) {
		operation_mode &= ~HT_INFO_OPERATION_MODE_OP_MODE_MASK;
		operation_mode |= new_op_mode;
		op_mode_changes++;
	}

	wpa_printf(MSG_DEBUG, "%s new operation mode=0x%X changes=%d\n",
		__func__, operation_mode, op_mode_changes);
	ht_add_info->data.operation_mode = host_to_le16(operation_mode);

	return op_mode_changes;
}

void debug_dump_ht_capability(struct ieee80211_ht_capability *ht_capability,
			      const char *msg)
{
	int i;

	printf("%s %s\n", __FUNCTION__, msg != NULL ? msg : "");

	if (!ht_capability) {
		printf("ht_capability is NULL\n");
		return;
	}
	printf("struct ieee80211_ht_capability {\n"
        	"\t__le16 capabilities_info=0x%X;\n"
        	"\tu8 mac_ht_params_info=0x%X;\n",
		le_to_host16(ht_capability->capabilities_info),
		ht_capability->mac_ht_params_info);
	printf("\tu8 supported_mcs_set[16]=");
	for (i = 0; i < 16; i++)
		printf(" 0x%X", ht_capability->supported_mcs_set[i]);
	printf("\n");

        printf("\t__le16 extended_ht_capability_info=0x%X;\n"
        	"\t__le32 tx_BF_capability_info=0x%X;\n"
        	"\tu8 antenna_selection_info=0x%X;\n",
		le_to_host16(ht_capability->extended_ht_capability_info),
		le_to_host32(ht_capability->tx_BF_capability_info),
		ht_capability->antenna_selection_info);
	printf("}\n");
}

void debug_dump_ht_additional_info(
		struct ieee80211_ht_additional_info *ht_add_info,
		const char *msg)
{
	int i;
	
	printf("%s %s\n", __FUNCTION__, msg != NULL ? msg : "");
	
	if (!ht_add_info) {
		printf("ht_add_info is NULL\n");
		return;
	}

	printf("struct ieee80211_ht_additional_info {\n"
        	"\tu8 control_chan=0x%X;\n"
		"\tu8 ht_param=0x%X;\n"
        	"\t__le16 operation_mode=0x%X;\n"
        	"\t__le16 stbc_param=0x%X;\n",
		ht_add_info->control_chan,
		ht_add_info->ht_param,
		le_to_host16(ht_add_info->operation_mode),
		le_to_host16(ht_add_info->stbc_param));

	printf("\tu8 basic_set[16]=");
	for (i = 0; i < 16; i++)
		printf(" 0x%0X", ht_add_info->basic_set[i]);
	printf("}\n");
}
