/*
 * hostapd / WDS initialization
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */
#include <string.h>
#include "includes.h"
#include "eloop.h"
#include "hostapd.h"
#include "driver.h"
#include "wds.h"

#define WDS_EXP_TIME	10

static int wds_dynamic_add(struct hostapd_data *hapd)
{
	int i;
	struct hostapd_bss_config *conf = hapd->conf;
	char *iface;

	for (i = 0; i < conf->wds_peer_num; i++) {
		iface = conf->wds_peer[i].iface;
		if (iface[0] == '\0')
			snprintf(iface, IFNAMSIZ, "%swds%d", conf->iface, i);
		if (hostapd_if_add(hapd, HOSTAPD_IF_WDS,
				iface, conf->wds_peer[i].peer_addr)) {
			if (errno != EEXIST) {
				printf("Could not add WDS iface: %s: %s\n",
					iface, strerror(errno));
				return -1;
			}
		}
	}

	return 0;
}

static void wds_dynamic_remove(struct hostapd_data *hapd)
{
	int i;
	struct hostapd_bss_config *conf = hapd->conf;
	char *iface;

	for (i = 0; i < conf->wds_peer_num; i++) {
		iface = conf->wds_peer[i].iface;
		if (hostapd_if_remove(hapd, HOSTAPD_IF_WDS,
			iface, conf->wds_peer[i].peer_addr)) {
			printf("Could not remove WDS iface: %s: %s\n",
				iface, strerror(errno));
		}
	}
}

static void wds_conn_expiration(struct hostapd_data *hapd,
				 struct hostapd_wds *wds)
{
	printf("WDS connection expired: " MACSTR "\n", MAC2STR(wds->peer_addr));

	hostapd_sta_set_flags(hapd, wds->peer_addr,
		0, ~(WLAN_STA_AUTH | WLAN_STA_ASSOC | WLAN_STA_AUTHORIZED));
	wds->flags &= ~WDS_PEER_FLG_CONNECTED;
}

static void wds_timer(void *eloop_ctx, void *timeout_ctx)
{
	struct hostapd_data *hapd = eloop_ctx;
	time_t now;
	int i;
	struct hostapd_wds *wds;

	eloop_register_timeout(10, 0, wds_timer, hapd, NULL);

	time(&now);

	for (i = 0; i < hapd->conf->wds_peer_num; i++) {
		wds = &hapd->conf->wds_peer[i];
		if ((wds->flags & WDS_PEER_FLG_CONNECTED) &&
		    now > (wds->last_beacon + WDS_EXP_TIME))
			wds_conn_expiration(hapd, wds);
	}
}

int wds_init(struct hostapd_data *hapd)
{
	int rc = 0;

	if (hapd->conf->wds_peer_num) {
		eloop_register_timeout(10, 0, wds_timer, hapd, NULL);
		rc = wds_dynamic_add(hapd);
	}

	return rc;
}

void wds_deinit(struct hostapd_data *hapd)
{
	if (hapd->conf->wds_peer_num) {
		eloop_cancel_timeout(wds_timer, hapd, NULL);
		wds_dynamic_remove(hapd);
	}
}

int wds_reconfig(struct hostapd_data *hapd, struct hostapd_config *oldconf,
	struct hostapd_bss_config *oldbss)
{
	wds_dynamic_remove(hapd);
	if (wds_dynamic_add(hapd))
		return -1;

	return 0;
}

static int wds_sta_update( struct hostapd_data *hapd,
			struct hostapd_wds *wds,
			u16 capability,
			struct ieee802_11_elems *elems)
{
	u8 rates[WLAN_SUPP_RATES_MAX];
	size_t rates_len = 0;
	struct ht_cap_ie ht_cap = {0};
	int flags = 0;

	if (elems->ht_capabilities &&
	    elems->ht_capabilities_len ==
		sizeof(struct ieee80211_ht_capability)) {
		ht_cap.id = 45;
		ht_cap.length = elems->ht_capabilities_len;
		memcpy(&ht_cap.data, elems->ht_capabilities,
			elems->ht_capabilities_len);
	}

	if (elems->supp_rates && elems->supp_rates_len <= sizeof(rates)) {
		memcpy(rates, elems->supp_rates, elems->supp_rates_len);
		rates_len += elems->supp_rates_len;
	}

	if (elems->ext_supp_rates &&
	    (elems->ext_supp_rates_len <= sizeof(rates) - rates_len)) {
		memcpy(rates + rates_len, elems->ext_supp_rates,
			elems->ext_supp_rates_len);
		rates_len += elems->ext_supp_rates_len;
	}

	return hostapd_sta_add(wds->iface, hapd, wds->peer_addr, 0,
		capability, rates, rates_len,
		elems->ht_capabilities ? &ht_cap : NULL, flags);
}

static void wds_connect(struct hostapd_data *hapd,
			u16 capability,
			struct ieee802_11_elems *elems,
			struct hostapd_wds *wds)
{
	if ( wds_sta_update(hapd, wds, capability, elems) ) {
			printf("Failed to update WDS sta, probably "
				"WDS network interface is down\n");
			return;
	}

	if (hostapd_sta_set_flags(hapd, wds->peer_addr,
		WLAN_STA_AUTHORIZED | WLAN_STA_WDS, (~0))) {
		printf("Failed to set WDS peer flags\n");
		return;
	}

	wds->flags |= WDS_PEER_FLG_CONNECTED;
}

void wds_process_beacon(struct hostapd_data *hapd,
			struct ieee80211_mgmt *mgmt,
			struct ieee802_11_elems *elems)
{
	int i;
	struct hostapd_bss_config *conf = hapd->conf;
	struct hostapd_wds *wds;
	time_t now;

	time(&now);

	for (i = 0; i < conf->wds_peer_num; i++) {
		wds = &conf->wds_peer[i];

		if (memcmp(mgmt->sa, wds->peer_addr, ETH_ALEN) == 0) {
			wds->last_beacon = now;
			if (!(wds->flags & WDS_PEER_FLG_CONNECTED))
				wds_connect(hapd,
					le_to_host16(mgmt->u.beacon.capab_info),
					elems, wds);
		}
	}
}

