#include "includes.h"

#include "common.h"
#include "eap_server/eap_i.h"
#include "eap_wsc.h"

#include "../../hostapd/udplib.h"

#define WSC_EAP_UDP_PORT	37000
#define WSC_EAP_UDP_ADDR	"127.0.0.1"
#define WSC_EAP_READ_TIMEOUT	5

static u8 *eap_wsc_com_buildReq(struct eap_sm *sm, 
				struct eap_wsc_data *data, 
				int id, 
				size_t *reqDataLen)
{
	u8 *req = NULL;
	int recvBytes;
	struct sockaddr_in from;
	struct sockaddr_in to;
	WSC_NOTIFY_DATA notifyData;
	WSC_NOTIFY_DATA *recvNotify;

	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	notifyData.type = WSC_NOTIFY_TYPE_BUILD_REQ;
	notifyData.u.bldReq.id = id;
	notifyData.u.bldReq.state = data->state;
	notifyData.length = 0;

	to.sin_addr.s_addr = inet_addr(WSC_EAP_UDP_ADDR);
	to.sin_family = AF_INET;
	to.sin_port = htons(WSC_EAP_UDP_PORT);

	if (udp_write(data->udpFdEap, (char *) &notifyData, sizeof(WSC_NOTIFY_DATA), &to) < 
		sizeof(WSC_NOTIFY_DATA)) {
		wpa_printf(MSG_INFO, "%s sending EAP message to upper layer failed", __FUNCTION__);
		data->state = FAILURE;
		return NULL;
	}
	
	recvBytes = udp_read_timed(data->udpFdEap, (char*) data->recvBuf, 
				   WSC_RECVBUF_SIZE, &from, WSC_EAP_READ_TIMEOUT);	

	if (recvBytes == -1) {
		wpa_printf(MSG_INFO, "%s reading EAP message from the upper layer failed", 
			__FUNCTION__);
		data->state = FAILURE;
		return NULL;
	}
	recvNotify = (WSC_NOTIFY_DATA*) data->recvBuf;

	if ((recvNotify->type != WSC_NOTIFY_TYPE_BUILD_REQ_RESULT) ||
	    (recvNotify->length == 0) ||
	    (recvNotify->u.bldReqResult.result != WSC_NOTIFY_RESULT_SUCCESS)) {
		wpa_printf(MSG_INFO, "%s build request failed somewhere", __FUNCTION__);
		data->state = FAILURE;
		return NULL;
	} 

	req = (u8*) malloc(recvNotify->length);
	if (!req) {
		wpa_printf(MSG_INFO, "%s memory allocation for request failed", __FUNCTION__);
		data->state = FAILURE;
		return NULL;
	}

	memcpy(req, recvNotify + 1, recvNotify->length);
	*reqDataLen = recvNotify->length;
	data->state = CONTINUE;

	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);
	return req;
}

static int eap_wsc_com_process(struct eap_sm *sm, struct eap_wsc_data *data,
			       u8 *respData, unsigned int respDataLen)
{
	int recvBytes;
	struct sockaddr_in from;
	struct sockaddr_in to;
	u8 *sendBuf;
	u32 sendBufLen;
	WSC_NOTIFY_DATA notifyData;
	WSC_NOTIFY_DATA *recvNotify;

	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	notifyData.type = WSC_NOTIDY_TYPE_PROCESS_RESP;
	notifyData.length = respDataLen;
	notifyData.u.process.state = data->state;

	sendBuf = (u8*) malloc(sizeof(WSC_NOTIFY_DATA) + respDataLen);
	if (!sendBuf) {
		wpa_printf(MSG_INFO, "%s memory allocation for sendBuf failed", __FUNCTION__);
		data->state = FAILURE;
		return -1;
	}
	
	memcpy(sendBuf, &notifyData, sizeof(WSC_NOTIFY_DATA));
	memcpy(sendBuf + sizeof(WSC_NOTIFY_DATA), respData, respDataLen);
	sendBufLen = sizeof(WSC_NOTIFY_DATA) + respDataLen;

	to.sin_addr.s_addr = inet_addr(WSC_EAP_UDP_ADDR);
	to.sin_family = AF_INET;
	to.sin_port = htons(WSC_EAP_UDP_PORT);

	if (udp_write(data->udpFdEap, (char*) sendBuf, sendBufLen, &to) < sendBufLen) {
		wpa_printf(MSG_INFO, "%s sending message to upper layer failed", __FUNCTION__);
		data->state = FAILURE;
		return  -1;
	}

	free(sendBuf);

	recvBytes = udp_read_timed(data->udpFdEap, (char*) data->recvBuf, 
				   WSC_RECVBUF_SIZE, &from, WSC_EAP_READ_TIMEOUT);

	if (recvBytes == -1) {
		wpa_printf(MSG_INFO, "%s Reading EAP message from upper layer failed", __FUNCTION__);
		data->state = FAILURE;
		return -1;
	}

	recvNotify = (WSC_NOTIFY_DATA*) data->recvBuf;
	wpa_printf(MSG_DEBUG, "%s recvNotify: type=%d, length=%d, result=%d", __FUNCTION__,
				recvNotify->type, 
				recvNotify->length, 
				recvNotify->u.processResult.result); 

	if ((recvNotify->type != WSC_NOTIFY_TYPE_PROCESS_RESULT) ||
	    (recvNotify->u.processResult.result != WSC_NOTIFY_RESULT_SUCCESS)) {
		wpa_printf(MSG_INFO, "%s process message failed somewhere", __FUNCTION__);
		data->state = FAILURE;
		return -1;
	}
	
	data->state =  CONTINUE;
	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);

	return 0;
}

	
static void *eap_wsc_init(struct eap_sm *sm)
{
	struct eap_wsc_data *data;

	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	data = (struct eap_wsc_data*)malloc(sizeof(struct eap_wsc_data));
	if (data == NULL) 
		return data;

	memset(data, 0, sizeof(*data));
	data->state = START;

	data->udpFdEap = udp_open();

	sm->eap_method_priv = data;
	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);
	return data;
}

static void eap_wsc_reset(struct eap_sm *sm, void *priv)
{
	struct eap_wsc_data *data = (struct eap_wsc_data *)priv;
	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	if (data == NULL)
		return;

	if (data->udpFdEap != -1) {
		udp_close(data->udpFdEap);
		data->udpFdEap = -1;
	}

	free(data);
	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);
}

static u8 *eap_wsc_buildReq(struct eap_sm *sm, void *priv, int id,
                            size_t *reqDataLen)
{
	u8 *ret = NULL;
	struct eap_wsc_data *data = (struct eap_wsc_data *)priv;
	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	switch (data->state) {
		case START:
		case CONTINUE:
			ret = eap_wsc_com_buildReq(sm, data, id, reqDataLen);
		default:
			wpa_printf(MSG_DEBUG, "%s: unexpedted state %d", __FUNCTION__, data->state);
			break;
	}
	wpa_printf(MSG_DEBUG, "%s << ret = 0x%X", __FUNCTION__, (unsigned int)ret);
	return ret;
}

static Boolean eap_wsc_check(struct eap_sm *sm, void *priv,
                             u8 *respData, size_t respDataLen)
{
	struct eap_hdr *resp = (struct eap_hdr*)respData;
	size_t len;

	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	if ((respDataLen < sizeof(*resp) + 2) ||
	    (len = ntohs(resp->length)) > respDataLen) {
		wpa_printf(MSG_INFO, "%s Invalid frame", __FUNCTION__);
		return TRUE;
	}
	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);
	return FALSE;
}

static void eap_wsc_process(struct eap_sm *sm, void *priv,
                                 u8 *respData, size_t respDataLen)
{
	struct eap_wsc_data *data = priv; 
	wpa_printf(MSG_DEBUG, "%s >> received packet (len=%d)", __FUNCTION__, respDataLen);
	
	if (eap_wsc_com_process(sm, data, respData, respDataLen) < 0) {
		wpa_printf(MSG_INFO, "%s WSC processing failed", __FUNCTION__);
		data->state = FAILURE;
	}
	wpa_printf(MSG_DEBUG, "%s <<", __FUNCTION__);
}

static Boolean eap_wsc_isDone(struct eap_sm *sm, void *priv)
{
	struct eap_wsc_data *data = priv;
	Boolean ret;
	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	ret = data->state == SUCCESS || data->state == FAILURE;
	wpa_printf(MSG_DEBUG, "%s << ret = %d", __FUNCTION__, ret);
	return ret;
}

static Boolean eap_wsc_isSuccess(struct eap_sm *sm, void *priv)
{
	struct eap_wsc_data *data = priv;
	Boolean ret;
	wpa_printf(MSG_DEBUG, "%s >>", __FUNCTION__);
	ret = data->state == SUCCESS;
	wpa_printf(MSG_DEBUG, "%s << ret = %d", __FUNCTION__, ret);
	return ret;
}

int eap_server_wsc_register(void)
{
        struct eap_method *eap;
        int ret;

        eap = eap_server_method_alloc(EAP_SERVER_METHOD_INTERFACE_VERSION,
                                      EAP_VENDOR_IETF, EAP_TYPE_WSC, "WSC");
        if (eap == NULL)
                return -1;

        eap->init = eap_wsc_init;
        eap->reset = eap_wsc_reset;
        eap->buildReq = eap_wsc_buildReq;
        eap->check = eap_wsc_check;
        eap->process = eap_wsc_process;
        eap->isDone = eap_wsc_isDone;
        eap->getKey = NULL;
        eap->isSuccess = eap_wsc_isSuccess;
        eap->get_emsk = NULL;

        ret = eap_server_method_register(eap);
        if (ret)
                eap_server_method_free(eap);
        return ret;
}


